// Rotational Inverted Pendulum Rev C
// Firmware for implementing a rotaional inverted pendulum on a PIC18F2431
// RS232 communciaitons are impleneted to facilitate testing and tuning and uses the EUSART peripheral
// IR communciations provide a remote control without a PC and use the capture and compare and timer 1 peripheral
// A potentiometer measures via the A/D 0 the position of the rotating arm in the horizontal plane
// A encoder with output in quadrature and zero position index are connected to the PIC QEI peripheral  
// Timer 0 provides main loop timing
// The motor control peripheral uses a PWM output to control the motor

#include <htc.h>
#include <conio.h>
#include <stdlib.h>
#include <math.h>

__CONFIG(1,ECIO); 							// External clock, RA6 is IO
__CONFIG(2,WDTDIS); 							// Disable watchdog

// Global registers are used to exchange data between interrupt service routines and main routine. 
unsigned char	rc_buffer[32]; 						// Stores data received via RS232 intterupts - circular buffer
unsigned char 	rc_head=0; 						// Pointer to last received data elements
unsigned char	ir_bits = 0; 						// While receiving an IR packet, holds the number of bits
unsigned long   ir_buffer=0; 						// Bit register that stores IR data while it's being received
unsigned char	ir_data=0; 						// Stores the resultant IR command if address and data quality check out


// Pin 1 - MCLR - Vdd - Held high
// Pin 2 - AN0 - A/D input used to measure motor shaft position
// Pin 3 - RA1 - Vdd - Held high
// Pin 4 - INDX - Takes pulse from encoder and zeros poition counter
// Pin 5 - QEA - Takes A pulses from encoder 
// Pin 6 - QEB - Takes B pulses from encoder 
// Pin 7 - Vdd - 5V
// Pin 8 - Vss - 0V
// Pin 9 - OSC1 - Input for 16 MHz oscillator
// Pin 10 - RA6 - Unconnected
// Pin 11 - RC0 - Unconnected
// Pin 12 - CCP2 - Falling edge input for IR recpetion
// Pin 13 - CCP1 - Rising edge input for IR reception
// Pin 14 - RC3 - Unconnected


// Pin 28 - RB7 - H-Bridge Driver disable
// Pin 27 - RB6 - Motor direction
// Pin 26 - RB5 - Unconnected
// Pin 25 - RB4 - Unconnected
// Pin 24 - RB3 - Unconnected
// Pin 23 - RB2 - Unconnected
// Pin 22 - RB1 - Unconnected
// Pin 21 - PWM0 - Speed Control to H-Bridge driver
// Pin 20 - Vdd - 5V
// Pin 19 - Vss - 0V
// Pin 18 - RX - RS232 receive from MAX232
// Pin 17 - TX - RS232 transmit to MAX232
// Pin 16 - RC5 - Unconnected
// Pin 15 - RC4 - Unconnected

// Initialization routine
void init(void)
{
	// Oscillator
	OSCCON = 0x00; 						// Use external oscillator

	// Ports
	PORTA = 0x00; 						// Just clear registers 
	TRISA = 0x1F; 						// RA0-RA4 are inputs
	PORTB = 0x00; 						// Initialize H-bridge drive to zero, and disable the drive, direction doesn't matter
	TRISB = 0x00; 						// RB0,6,7 are all outputs.  Others are unconnected
	PORTC = 0x00; 						// Just clear registers 
	TRISC = 0x86; 						// RC2,RC1 are inputs for capture and compare, RC7 for RS232 receive 


	// Use timer 0 for main program loop timing
	T0CON = 0x80; 						// 16 MHz Oscill - Divide by 4 - then prescale by 2 - 2 MHz

	// Use timer 1 for infra-red remote timing
	T1CON = 0xB0; 						// 16 MHz Oscill - Divide by 4 - then prescale by 8 - 500 kHz 
	TMR1IF = 0; 						// Clear any pending events

	// Setup RS232 communications
	TXSTA = 0xA0;						// Transmitter based on low BRG, 8 bit async,
	RCSTA = 0x90; 						// Enable 8 bit continuous recpetion, 
	BAUDCTL = 0x00; 					// 8 bit baud rate 
	SPBRG	= 0x0c; 					// 19.2 kb @16MHz

	// Setup PWM Port
	PTCON0 = 0x00; 						// 4 MHz free running mode
	PTCON1 = 0x80; 						// Enable the PWM counter to count up
	PWMCON0 = 0x21; 					// PWM 0 and 1 are active in independant mode
	PWMCON1 = 0x00; 					// Update the duty timer immediately
	PDC0L = 0x00; 						// Zero duty cycle on start up
	PDC0H = 0x00; 						// Zero duty cycle on start up
	PTPERH = 0x00; 						// PWM time base
	PTPERL = 0xFF; 						// PWM Time base 15,625 Hz

	// Port A setup position encoder
	ANSEL0 = 0x01; 						// AN0 is set to analog input mode 


	// Set up encoder peripheral
	QEICON = 0x94; 						// INDX used to reset x4 update mode
	MAXCNTH = 0x0F; 					// Max count of 4095
	MAXCNTL = 0xFF; 					// Max count of 4095

	// Set up capture modules
	CCP1CON = 0x05; 					// Set CCP1 as a rising edge interrupt
	CCP2CON = 0x04; 					// Set CCP2 as a falling edge interrupt


	// Setup A/D converter to measure horizontal position
	// SQ mode 1
	// no FIFO, right justified and focs/2
	ADCON0 = 0x01;						// Enable single shot converiosn
	ADCON2 = 0x80;						// Right justify

	GIE = 1; 						// Global interrupt enable
	PEIE = 1; 						// Enable peripheral interrupts
	RCIE = 1; 						// Enable receive channel interrupt
	RCIF = 0; 						// Clear any pending events
	CCP1IE = 1; 						// Enable interrupts on CCP1
	CCP1IF = 0; 						// Clear any pending events
	CCP2IE = 1; 						// Enable interrupts on CCP2
	CCP2IF = 0; 						// Clear any pending events


}


void main(void)
{

	unsigned char	i,j,k;					//Generic Looping Variable
	signed int	horizontal_position;			// The positon of the rotating arm horizontal angle
	unsigned char	horz_l,horz_h;				// High and low bytes from A/D register -512 to 512
	signed int	horizontal_setpoint = 0;		// Initialize set point to center position
	signed int	new_horz_setpoint = 0;			// Stores a user desired setpoint - used for ramping of horz setpoint
	signed int	vertical_position;			// The positon of the pendulum arm relative to encoder zero
	unsigned char	vert_l,vert_h;				// High and low bytes of POSCNT the QEI register
	unsigned char	dir = 0;				// 0-CCW 1-CW from the QEI peripheral
	signed int	vertical_setpoint = 0;			// Seems obvious but this register initialized
	signed int	error[2][3] = {{0,0,0},{0,0,0}};	// Errors from the last three time steps for vertical and horizontal controllers
	unsigned int	abs_error;				// Error magnitude around the vertical set point
	signed long	outdrv[2][2] = {{0,0},{0,0}};		// For both controllers, the most recent and present output signal
	signed long	driver = 0;					// Resultant output to be sent to the H-Bridge drivers
	signed long	term[2][3] = {{0,0,0},{0,0,0}};		// Stores, for each controller, the value of each term
	unsigned char	serial_op_cnt=0;			// Counts main loops, used to calcualte 1/2 second to send out RS232 packets
	unsigned int 	cont_mode=4;				// 4 - Stop motor
								// 2 - Upright control mode
								// 1 - Swingup control mode
	unsigned int 	temp_pwm;				// Unsigned int for output to PWM
	signed int	kpid[2][3] = {{0,0,0},{0,0,0}};		// Stores the tuning paramters for each controller
	signed int	vert_set_avg = 789;			// The average encoder reading over the travel of the rotating arm
	signed int	set_select;				// The index into the motor wobble correction values 
	signed int	set_interp;				// The interpolated deviation from the average vertical set point 
	signed int	vert_offset[32] = {23,14,12,9,5,3,0,-3, // Equally spaced samples, on the travel of the rotating arm, expressed
					-5,-8,-10,-11,-12,-13,-13,-14,	// as deviations from the average value
					-14,-13,-12,-10,-9,-7,-4,-1,
					1,5,8,11,15,18,21,22};

	// Run the initialization routine to set up all registers
	init();


	// Hard coded initialization parameters, this allows the RIP to run without the WIndows interface		
	kpid[0][0] = 0x0634; // Horizontal - Kp - 3
	kpid[0][1] = 0xF69B; // Horizontal - Ki - 0.5
	kpid[0][2] = 0x0333; // Horizontal - Kd - 0.05
	kpid[1][0] = 0x11F5; // Veritcal Kp - 5
	kpid[1][1] = 0xE18E; // Vertical Ki - 20
	kpid[1][2] = 0x0CCD; // Vertical Kd - 0.2

	// Main loop
	while(1)
	{
		GODONE = 1;					// Start the next A/D conversion
		horz_l = ADRESL;				// Store the lower and upper value of A/D
		horz_h = ADRESH;				// 10 bit measurment of pot voltage divider
		vert_l = POSCNTL;				// Store the lower and upper value of the QEI counter
		vert_h = POSCNTH;				// 12 bit measurement of encoer position
		if(QEICON&0x20)					// Direction of vertical encoder motion
			dir = 0x00;				// Counter Clock Wise
		else
			dir = 0x01;				// Clock Wise



		// Determine horizontal position and error
		horizontal_position = (horz_l+horz_h*256)-512;	// Horizontal position is 10 bit - scale to 512 to -512
		if((horizontal_position > 450) || (horizontal_position  < -450) )
			cont_mode = 4;				// Stop the motor if out of bounds
		error[0][2] = error[0][1];			// PID calcuation requies present error and two most recent error values
		error[0][1] = error[0][0];
		error[0][0] = horizontal_position - horizontal_setpoint;
								// The present error is the difference in the rotating arms position from the
								// user requested set point
		
		// The motor has a wobble that results in a slight twisting of the rotating arm
		// The twisting is repeatable, from measurments the average 'up' encoder position is caluted and hard coded
		// 32 evenly spaced samples are taken of the deviation from this average value
		// For the present horizontal position the vertical 'up' is interpolated from the samples

		set_select = ((horizontal_position + 512)>>5);	// The index that is less than and nearest to present horizontal position
		set_interp = ((vert_offset[set_select+1]-vert_offset[set_select]) * ((horizontal_position + 512)&0x001F))>>5;
								// Interpolate between adjacent samples of the vertical deviation
		vertical_setpoint = vert_set_avg + set_interp+vert_offset[set_select];
								// Calculate the up value using the deviation


		// Determine vertical position and error
		vertical_position = (vert_l + vert_h*256); 	// Vertical position is 12 bit - scale to 0 to 4095
		error[1][2] = error[1][1];			// PID calcuation requies present error and two most recent error values
		error[1][1] = error[1][0];
		error[1][0] = vertical_position-vertical_setpoint;
								// The present error is the difference between the measured osition and the
								// estimated 'up' position
		if(error[1][0] > 2047)				// For the swingup mode and the switch to the upright control mode it is 
			error[1][0] -= 4096;			// necessary to know the magnitude deviation from the 'up' position.  So the 
		abs_error = abs(error[1][0]);			// error must be expressed -2047 to 2048 and then as a magnitude


		// Swing Up control mode
		if (cont_mode == 1)
		{
			// If the pendulum is within 5.5 Deg, transition to upright control mode
			if(abs_error < 64 )			 
			{
				outdrv[1][0] = 0;		// Stop driving the motor
				horizontal_setpoint = horizontal_position;
				new_horz_setpoint = horizontal_setpoint;
								// Set the horizontal set point to the current horizontal position
								// this will hopefully prevent a large step change while the pendulum 
								// initially trying to achieve stability 

				error[0][2] = 0;		// Zero the previous errors before starting the new controllers
				error[0][1] = 0;
				error[0][0] = 0;
				cont_mode=2;			// Change to upright control mode
			}


			// In swing up mode

			// Drive at a constant value - Magic number that works well
			outdrv[1][0] = abs_error<<4;

			// If above the horizon of the encoder and
			// approaching the up position then move the base under the pendulum arm
			// else if falling move away to increase energy for next swing
			// If below the horizon swing in the direction
			if	(abs_error < 1024)
			{	
				if(!dir)
					outdrv[1][0] = -outdrv[1][0];
		
			}
			else
			{	
				if(dir)
					outdrv[1][0] = -outdrv[1][0];
		
			}
 
			outdrv[0][0] = 0;			// In swing up mode the vertical controller has no output

		}	
		
		// Upright control mode
		else if (cont_mode == 2)
		{
			// Controlling in the upright position
			// If we have fallen out of contralabilty range kill the motor
			if(abs_error > 384)
				cont_mode=4;

			// Calcualte the vertical controller output using tuning variables,
			// two previous erros and the past output value	
			term[1][0] = (signed long)error[1][0] * (signed long)kpid[1][0];
			term[1][1] = (signed long)error[1][1] * (signed long)kpid[1][1];
			term[1][2] = (signed long)error[1][2] * (signed long)kpid[1][2];
			outdrv[1][0] = outdrv[1][1] + term[1][0] + term[1][1] + term[1][2]; 
			outdrv[1][1] = outdrv[1][0];

			// Ramp up setpoint changes
			if(new_horz_setpoint != horizontal_setpoint)				
			{	
				// Using serial loop timer, slowly ramp actual set point to that requested by the user
				if(serial_op_cnt & 0x04)
				{
					if(horizontal_setpoint>new_horz_setpoint)
						horizontal_setpoint--;
					if(horizontal_setpoint<new_horz_setpoint)
						horizontal_setpoint++;
				}
			}
			// Calcualte the horizontal controller output using tuning variables,
			// two previous erros and the past output value	
			term[0][0] = (signed long)error[0][0] * (signed long)kpid[0][0];
			term[0][1] = (signed long)error[0][1] * (signed long)kpid[0][1];
			term[0][2] = (signed long)error[0][2] * (signed long)kpid[0][2];
			outdrv[0][0] = outdrv[0][1] + term[0][0] + term[0][1] + term[0][2]; 
			outdrv[0][1] = outdrv[0][0];

		}	

		// If in control mode 1,2 were are still controlling the motor
		if(!(cont_mode==4))
		{
			
			PORTB |= 0x80;				// Enable hte H-Bridge drivers		 
												

			driver = outdrv[0][0] - outdrv[1][0] ;
								// The sense of the horizontal and vertical
								// controllers is opposite
			
			// Based on sign of output driver, switch the
			// direction of the motor
	    		if (driver > 0)
				PORTB |= 0x40;
			else
				PORTB &= 0xBF;
			
			// Calcualte the magnitude of the output
			if(driver < 0)
				driver = -driver;

			// Limit the driver output
			if (driver > 100000)				
				driver = 100000;
			
			temp_pwm = ((int)((driver)>>7));		// Convert PWM value to integer
			PDC0H = (temp_pwm >>8) & 0x00FF;		// Need comment
			PDC0L = temp_pwm & 0x00FF;
		}
		
		// If in mode 4 stop the controllers and disable the motor
		else
		{
			
			PORTB &= 0x7F;					 // Disable the driver
												
			outdrv[1][0] = 0;				 // Zero the driver and PWM command
			outdrv[1][1] = 0;
			outdrv[0][0] = 0;
			outdrv[0][1] = 0;
			PDC0H = 0;		
			PDC0L = 0;

		}

		
		////////////////////////////////////////////////////////////
		// Send out 13 byte Package to update display
		// For testting and tuning only
		// Only send it out every 1/2 second
		serial_op_cnt++;
		if(serial_op_cnt==32)
		{
			serial_op_cnt=0;
			while(!TRMT) 
				continue;
				TXREG = 0xAA;				// Header		
			while(!TRMT) 
				continue;
				TXREG = 0x55;				// Header
			while(!TRMT) 
				continue;
				TXREG = 0xAA;				// Header
			while(!TRMT) 
				continue;
				TXREG = vertical_position & 0x00FF;	// Low byte of vertical encoder position
			while(!TRMT) 
				continue;
				TXREG = (vertical_position>>8) & 0x00FF;// High byte of vertical encoder position
			while(!TRMT) 
				continue;
				TXREG = horizontal_position & 0x00FF;	// Low byte of horizontal pot. position
			while(!TRMT) 
				continue;
				TXREG = (horizontal_position>>8) & 0x00FF;		// High byte of horizontal pot. position
			while(!TRMT) 
				continue;
			if(QEICON&0x20)					// Direction of vertical encoder motion
				TXREG = 0x00;
			else
				TXREG = 0x01;
			while(!TRMT) 
				continue;
				TXREG = (outdrv[1][0]) & 0x00FF;	// Test data
			while(!TRMT) 
				continue;
				TXREG = ((outdrv[1][0])>>8) & 0x00FF;	// Test data
			while(!TRMT) 
				continue;
				TXREG = (outdrv[1][0]>>16) & 0x00FF;	// Test data 
			while(!TRMT) 
				continue;
				TXREG = ((outdrv[1][0])>>24) & 0x00FF;	// Test data 
			while(!TRMT) 
				continue;
				TXREG = 0;				// Tailer 

			
		}	
		////////////////////////////////////////////////////////////


		////////////////////////////////////////////////////////////
		// Wait until 1/64 of a second has passed
		while(!TMR0IF) 
		{
			/////////////////////////////////////////////////////////////
			// Check to see if we have received a complete message 
			// from the mothership
			// 0x55  
			// Controller Mode 
			// Horizontal Setpoint High Byte
			// Horizontal Setpoint Low Byte
			// Vertical Setpoint High Byte
			// Vertical Setpoint Low Byte
			// Horizontal Kp High Byte
			// Horizontal Kp Low Byte
			// Horizontal Ki High Byte
			// Horizontal Ki Low Byte
			// Horizontal Kd High Byte
			// Horizontal Kd Low Byte
			// Vertical Kp High Byte
			// Vertical Kp Low Byte
			// Vertical Ki High Byte
			// Vertical Ki Low Byte
			// Vertical Kd High Byte
			// Vertical Kd Low Byte
			// 13 zero bytes
  			// 0xAA

			// Check the circular buffer (32 bytes) to see there is a valid packet in it
			i=0;
			while(!(i & 0xE0))
			{
				// Valid packet will be headed with a ox55 and tailed by a 0xAA and be exactly 32 bytes long
				if ((rc_buffer[i]==0x55) && (rc_buffer[(i+31)&0x1f]==0xAA))
				{
					// Set 0x55 byte to zero to show that this package have been read				
					rc_buffer[i] = 0;				
					rc_buffer[(i+31)&0x1f] = 0;
					i++; i = i & 0x1f;

					// Command
					cont_mode = rc_buffer[i];
					rc_buffer[i] = 0;
					i++; i = i & 0x1f;
					
					// Setpoints are stored as two bytes each
					new_horz_setpoint = rc_buffer[i];
					new_horz_setpoint <<= 8;
					i++; i = i & 0x1f;
					new_horz_setpoint |= rc_buffer[i];
					i++; i = i & 0x1f;


					vertical_setpoint = rc_buffer[i];
					vertical_setpoint <<= 8;
					i++; i = i & 0x1f;
					vertical_setpoint |= rc_buffer[i];
					i++; i = i & 0x1f;
					
					// Tuning parameters
					for(j=0;j<2;j++)
					{
						for(k=0;k<3;k++)
						{
							kpid[j][k] = rc_buffer[i];
							kpid[j][k] <<= 8;
							i++; i = i & 0x1f;
							kpid[j][k] |= rc_buffer[i];
							i++; i = i & 0x1f;
						}
					}
				}
				
				i++;

			
				// Handle IR commands
				switch(ir_data)
				{
					// Play button - goto swing up mode
					case 0x16:
						cont_mode = 1;
						break;
					// Stop button - Off mode
					case 0x17:
						cont_mode = 4;
						break;
					// Rewind button - decrease set point
					case 0x18:
						new_horz_setpoint = horizontal_setpoint - 25;
						//horizontal_setpoint -= 10;
						break;
					// Forward button - increase set point
					case 0x19:
						new_horz_setpoint = horizontal_setpoint + 25;
						//horizontal_setpoint += 10;
						break;


				}
				ir_data = 0;
			}					
			/////////////////////////////////////////////////////////////		
		}
		TMR0IF = 0;

		TMR0H = 0x85;	// Magic number 0x3D09 - 31250.  TMR0 counts at 1 MHz (including prescalar)
		TMR0L = 0xED;	// So 1/64 of a second is  counts, so if we start TMR 0 at (65535-31250)

		////////////////////////////////////////////////////////////
		
	}

}



// Interrupt routines
void interrupt low_isr(void)
{
	// RS232 data received
	if((RCIE)&&(RCIF))
	{
		rc_head++;						// Move to next byte in buffer			
		rc_head = rc_head & 0x1F;
		rc_buffer[rc_head]=RCREG;				// Store data
		RCIF = 0;						// Clear interrupt flag
	}


	// Handle rising edge on CPP from TSOP IR receiver
	if((CCP1IE)&&(CCP1IF))
	{
		TMR1L = 0;						// On every rising edge, restart
		TMR1H = 0;						// Timer 1.  This timer is used to 
		T1CON |= 0x01;						// to check if we have received a	
		CCP1IF = 0;						// one or a zero
	}

	// Handle falling edge on CCP from TSOP IR receiver
	if((CCP2IE)&&(CCP2IF))						
	{
		if((T1CON&0x01))					// The timer will be running if we have 
		{							// received a rising edge previously

			ir_bits++;					// Every falling edge, after the first is a data bit
			ir_buffer >>= 1;				// Shift data register over to make room for next bit
			if(CCPR2H&0xFE)					// If the captured timer value exceed 1ms then this is a one
				ir_buffer |= 0x80000000;		// otherwise it is zero

			
			if(ir_bits>33)					// There are 32 data bits.  The first falling edge (off the 4.5ms
			{						// pulse) is dropped out of the data register but is in the bit
				CCP1IE = 0;				// count
				CCP2IE = 0;				// After all data is received, enable the timer 1 interrupt and
				TMR1IE = 1;				// disable the CCP1 and CCP2 interrupts - all repeated data will be ignored
				TMR1IF = 0;
			}
			else
			{
				T1CON &= 0xFE;				// If more data is expected, clear the timer enable in expectation
			}						// of another rising edge

		}
		else
		{
			ir_buffer = 0;					// Initialize the storage register when we receive
			ir_bits = 0;					// the first falling edge
		}		

		CCP2IF = 0;	

	}

	// Timer 1 timeout interrupt.
	if((TMR1IE)&&(TMR1IF))
	{
		ir_bits = 0;						// After receiving 33 bits we've timed out (after about 4 secs)
		CCP1IE = 1;						// Restart the interrupts looking for rising and falling edges
		CCP2IE = 1;						// in expectation of more data
		TMR1IE = 0;						// Disable timer 1 interrupt
		CCP1IF = 0;
		CCP2IF = 0;
		if((ir_buffer&0x0000FFFF) == 0x00007B80)		// Confirm that expected address is correct
			ir_buffer = ir_buffer>>16;			
									// Confirm the high and low bytes of data are logical inverses
			if(!(((ir_buffer&0xFF00)>>8) & (ir_buffer&0x00FF)))
				ir_data = (unsigned char)(ir_buffer&0x00FF);
	}								// If pass command to main routine


	GIE = 1;
	PEIE = 1;
}
